# -*- coding: utf-8 -*-
"""
Utilities for general EDGAR website tasks.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""
import os
import re
import logging

from pyedgar import config
#  import FEED_ROOT, FEED_CACHE_ROOT, INDEX_ROOT, INDEX_CACHE_ROOT

__logger = logging.getLogger(__name__)

ACCESSION_RE = re.compile(r'(?P<accession>\d{10}-?\d\d-?\d{6})', re.I)

def get_filing_path(*args, **kwargs):
    """
    Return filepath to local copy of EDGAR filing.
    Filing document is .nc file with full submission, including main filing and exhibits/attachments.
    Does some sanity checking on top of config.get_filing_path

    :param args: Tries to guess cik/accession in the args passed in. cik/accession passed in by kwargs overrides these args.
    :param kwargs: dictionary to be passed to config.format_filing_path.

    :return: Full path to local filing document. Equal to ``join(FILING_ROOT, format_filing_path(**kwargs))``
    :rtype: string
    """
    cik = kwargs.get('cik', None)
    accession = kwargs.get('accession', None)

    if args:
        if accession is None:
            accession = [arg for arg in args if ACCESSION_RE.match(str(arg))]
            if accession:
                accession = accession[0]
            else:
                accession = None

        if cik is None:
            for arg in args:
                if len(str(arg)) <= 10:
                    try:
                        cik = int(arg)
                    except (ValueError, TypeError):
                        continue
                    break

    if cik is None and accession is None:
        raise ValueError("Requires non-missing CIK({}) or Accession({}). Got: args:{}, kwargs:{}"
                         .format(cik, accession, args, kwargs))

    try:
        clean_ac = ACCESSION_RE.search(accession).group('accession')
        if len(clean_ac) == 18: # no dashes found. Add dashes.
            clean_ac = clean_ac[:10] + '-' + clean_ac[10:12] + '-' + clean_ac[12:]
    except AttributeError: # no .group found.
        clean_ac = accession

    if cik is not None:
        kwargs['cik'] = cik
    if accession is not None:
        kwargs['accession'] = accession

    formatted_filename = config.format_filing_path(**kwargs)

    return os.path.join(config.FILING_ROOT, formatted_filename)

def walk_files(root_dir, filename_regex=None, return_dirs=False):
    """
    Iteratively walk directories and files, returning full paths.

    :param str root_dir: The root directory at which to start searching.
    :param re filename_regex: Regular expression (or string pattern) to which
              files or directories must match.
    :param bool return_dirs: Return directories as well as files.

    :return: Full path to filename or directory that matches optional regex.
    :rtype: string
    """
    if filename_regex is not None:
        try:
            # One can chain re.compile calls. This obviates checking type.
            filename_regex = re.compile(filename_regex)
        except (re.error, TypeError) as e:
            __logger.error("Regular expression provided is invalid!")
            raise(e)
    for r,ds,fs in os.walk(root_dir):
        if return_dirs:
            for d in ds:
                if filename_regex is not None and not filename_regex.search(d):
                    continue
                yield os.path.join(r,d)
        for f in fs:
            if filename_regex is not None and not filename_regex.search(f):
                continue
            yield os.path.join(r,f)



# -*- coding: utf-8 -*-
"""
Exceptions for the project.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

class InputTypeError(Exception):
    """Input to method is of wrong type."""

class WrongFormType(Exception):
    """Form provided is of wrong type"""

class NoFormTypeFound(Exception):
    """Form type not found in file (no <FORM-TYPE> tag)"""

class NoCIKFound(Exception):
    """CIK code not found in file (no <CIK> tag)"""

class EDGARFilingFormatError(Exception):
    """File does not appear to be in proper EDGAR SGML format."""



# -*- coding: utf-8 -*-
"""
Utilities for general EDGAR website and ftp tasks, including download functionality.

index URL: http://www.sec.gov/Archives/edgar/data/2098/0001026608-05-000015-index.htm
complete submission URL: http://www.sec.gov/Archives/edgar/data/2098/000102660805000015/0001026608-05-000015.txt
Exhibit/form URL: http://www.sec.gov/Archives/edgar/data/2098/000102660815000007/acu_10k123114.htm

EDGAR FTP change in 2016:
<2016 ftp URL: ftp://ftp.sec.gov/edgar/data/2098/0000002098-96-000003.txt
>2016 http URL: https://www.sec.gov/Archives/edgar/data/2098/0000002098-96-000003.txt

EDGAR HTML specification: https://www.sec.gov/info/edgar/ednews/edhtml.htm

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# Stdlib imports
import os
import re
import logging
import subprocess
import datetime as dt
from time import sleep

# 3rd party imports
import requests

# Module Imports
from pyedgar import config
from pyedgar import utilities

# Local logger
_logger = logging.getLogger(__name__)

# Constants
# Used to be (before FTP changed to AWS S3): ftp://ftp.sec.gov
EDGAR_ROOT = "https://www.sec.gov/Archives"
REQUEST_HEADERS = {
    "User-Agent": getattr(
        config, "USER_AGENT", "pyedgar downloader (fallback UA, shame) from gaulinmp+badpyedgarUA@gmail.com"
    )
}


def parse_url(url, url_re=re.compile(r"/(?P<cik>\d{1,10})/" r"(?P<accession>\d{10}-?\d\d-?\d{6})", re.I)):
    """Return CIK and Accession from an EDGAR HTTP or FTP url.

    Arguments:
        url (string): URL from EDGAR website or FTP server.
        url_re (Pattern): regular expression with cik and accession groups

    Returns:
        tuple: Tuple of strings in the form (cik, accession) or (None, None)
    """
    res = url_re.search(url)
    if res:
        acc = res.group("accession")
        if len(acc) == 18:
            acc = acc[:10] + "-" + acc[10:12] + "-" + acc[12:]
        return res.group("cik"), acc
    return None, None


def get_edgar_urls(cik, accession=None):
    """Generage URLs for EDGAR 'bulk download' and 'user facing' sites.
    `cik` parameter can be object with `cik` and `accession` attributes or keys.

    Arguments:
        cik (str,dict,object): String CIK, or object with cik and accession attributes or keys.
        accession (str): String ACCESSION number, or None if accession in CIK object.

    Returns:
        tuple: Tuple of strings in the form (raw url, Filing Index url)
    """
    _obj = utilities.get_cik_acc(cik, accession=accession)
    cik, accession = _obj.get("cik", None), _obj.get("accession", None)

    if cik is None or accession is None:
        return ("", "")

    return (
        "{}/edgar/data/{}/{}.txt".format(EDGAR_ROOT, cik, accession),
        "{}/edgar/data/{}/{}-index.htm".format(EDGAR_ROOT, cik, accession),
    )


def edgar_links(cik, accession=None, index=True, raw=True):
    """Generage HTML encoded links (using `a` tag) to EDGAR 'bulk download' and 'user facing' sites.
    `cik` parameter can be object with `cik` and `accession` attributes or keys.

    Arguments:
        cik (str,dict,object): String CIK, or object with cik and accession attributes or keys.
        accession (str): String ACCESSION number, or None if accession in CIK object.
        index (bool): Include link to Index of accession
        raw (bool): Include link to raw text of accession

    Returns:
        str: HTML `a` tag with href pointing to EDGAR index and/or raw file of of `accession`.
    """
    _raw, _idx = get_edgar_urls(cik, accession=accession)

    _html = []
    if index:
        _html.append("<a href='{}' target=_blank>Index</a>".format(_idx))
    if raw:
        _html.append("<a href='{}' target=_blank>Raw</a>".format(_raw))

    return " ".join(_html)


def get_feed_url(date):
    """Get URL path to daily feed gzip file.

    Arguments:
        date (datetime): Date for feed file

    Returns:
        str: URL to feed file on `date`"""
    return "{0}/edgar/Feed/{1:%Y}/QTR{2}/{1:%Y%m%d}.nc.tar.gz".format(EDGAR_ROOT, date, utilities.get_quarter(date))


def get_index_url(date_or_year, quarter=None, compressed=True):
    """
    Get URL path to quarterly index file.
    Do not feed it a year and no quarter.

    Arguments:
        date_or_year (datetime,int): Datetime from which the index's quarter will be calculated, or integer year
        quarter (int,None): Quarter of the index, will be inferred from date_or_year if None.
    """
    try:
        year = date_or_year.year
        # Leave this here, because if date_or_year isn't a datetime, we can't calculate the quarter. So error out.
        if quarter is None:
            quarter = utilities.get_quarter(date_or_year)
    except AttributeError:
        # Then date_or_year is an integer, cast just in case
        year = int(date_or_year)
        if quarter is None:
            raise AttributeError("Must either pass either datetime or both integer year AND quarter")

    ext = "gz" if compressed else "idx"

    return "{0}/edgar/full-index/{1}/QTR{2}/master.{3}".format(EDGAR_ROOT, year, quarter, ext)


def download_form_from_web(cik, accession=None):
    """
    Sometimes the cache file is not there, or you do not have local cache.
    In those cases, you can download the EDGAR forms from S3 directly.

    Arguments:
        cik (str,dict,object): String CIK, or object with cik and accession attributes or keys.
        accession (str): String ACCESSION number, or None if accession in CIK object.

    Returns:
        tuple: Tuple of strings in the form (bulk DL url, user website url)
    """
    _raw, _ = get_edgar_urls(cik, accession=accession)

    r = requests.get(_raw, headers=REQUEST_HEADERS)

    data = r.content

    for _decode_type, _errors in zip(("latin-1", "utf-8", "latin-1"), ("strict", "strict", "ignore")):
        try:
            return data.decode(_decode_type, errors=_errors)
        except (UnicodeDecodeError, ValueError):
            continue


def has_curl():
    try:
        subp = subprocess.run(["curl", '-A "pyedgar test"', "https://www.google.com"], capture_output=True)
    except FileNotFoundError:
        # apparently if curl isn't found it's FileNotFoundError.
        return False
    except Exception:
        return False
    return True


def download_from_edgar(
    edgar_url,
    local_path,
    overwrite=False,
    use_requests=False,
    chunk_size=10 * 1024 ** 2,
    overwrite_size_threshold=-1,
    sleep_after=0,
):
    """
    Generic downloader, uses curl by default unless use_requests=True is passed in.

    Arguments:
        edgar_url (str): URL of EDGAR resource.
        local_path (Path, str): Local path to write to
        overwrite (bool): Flag for whether to overwrite any existing file (default False).
        use_requests (bool): Flag for whether to use requests or curl (default False == curl).
        chunk_size (int): Size of chunks to write to disk while streaming from requests
        overwrite_size_threshold (int): Existing files smaller than this will be re-downloaded.
        sleep_after (int): Number of seconds to sleep after downloading file (default 0)

    Returns:
        (str, None): Returns path of downloaded file (or None if download failed).
    """
    if not os.path.exists(os.path.dirname(local_path)):
        raise FileNotFoundError("Trying to write to non-existant directory: {}".format(os.path.dirname(local_path)))

    if os.path.exists(local_path):
        loc_size = os.path.getsize(local_path)
        if not overwrite and loc_size > overwrite_size_threshold:
            _logger.info("Skipping cache file (%s bytes) at %r", "{:,d}".format(loc_size), local_path)
            return local_path

        _logger.warning("Removing existing file (%s bytes) at %r", "{:,d}".format(loc_size), local_path)
        os.remove(local_path)

    _useragent = REQUEST_HEADERS["User-Agent"]

    if not use_requests:
        _logger.debug('curl -A "%s" %s -o %s', _useragent, edgar_url, local_path)
        subp = subprocess.run(["curl", '-A "{}"'.format(_useragent), edgar_url, "-o", local_path], capture_output=True)
        _logger.debug(subp.stdout)
        sleep(sleep_after)
        if subp.returncode != 0:
            raise Exception("Error {} downloading with curl: {}".format(subp.returncode, subp.stderr))
    else:
        _logger.debug("requests.get(%r, headers=%r) >> %r", edgar_url, REQUEST_HEADERS, local_path)
        try:
            with requests.get(edgar_url, headers=REQUEST_HEADERS, stream=True) as response:
                expected_len = int(response.headers["content-length"])
                with open(local_path, "wb") as fh:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        if chunk:  # filter out keep-alive new chunks
                            fh.write(chunk)
            sleep(sleep_after)
        except Exception as excp:
            raise Exception("Error downloading with requests: {}".format(excp)) from excp

        # Now check what we downloaded was what we expected
        try:
            loc_size = os.path.getsize(local_path)
        except FileNotFoundError:
            raise Exception("Error downloading with requests to: {}".format(local_path))
        if expected_len != loc_size:
            _logger.exception("requests downloaded {:,d} bytes but expected {:,d}".format(loc_size, expected_len))

    if os.path.exists(local_path):
        _logger.info("Done downloading %.3f MB to %s", os.path.getsize(local_path) / 1024 ** 2, local_path)
        return local_path
    return None


def download_feed(date, overwrite=False, use_requests=False, overwrite_size_threshold=8 * 1024, sleep_after=0):
    """Download an edgar daily feed compressed file.

    Args:
        date (datetime, str): Date of feed file to download. Can be datetime
            or string (YYYYMMDD format with optional spacing).
        overwrite (bool): Flag for whether to overwrite any existing file (default False).
        use_requests (bool): Flag for whether to use requests or curl (default False == curl).
        overwrite_size_threshold (int): Existing files smaller than this will be re-downloaded.
        sleep_after (int): Number of seconds to sleep after downloading file (default 0)

    Returns:
        str: output file path
    """
    date = utilities.parse_date_input(date)

    if date.weekday() >= 5:  # skip sat/sun, because computers don't work weekends. Union rules, I think?
        return None

    feed_path = config.get_feed_cache_path(date)
    url = get_feed_url(date)

    return download_from_edgar(
        url,
        feed_path,
        overwrite=overwrite,
        use_requests=use_requests,
        overwrite_size_threshold=overwrite_size_threshold,
        sleep_after=sleep_after,
    )


def download_feeds_recursively(
    start_date, end_date=None, overwrite=False, use_requests=False, overwrite_size_threshold=8 * 1024, loop_sleep=1
):
    """Download edgar daily feed compressed files recursively from start to end.
    If `end_date` is `None`, default to today.
    Error downloads, or rate limited downloads are around 6kb,
    thus set the overwrite_size_threshold to larger than that
    to automatically re-download those error files.

    Args:
        start_date (datetime, str): Starting date of feeds to download. Can be datetime
            or string (YYYYMMDD format with optional spacing).
        end_date (None,datetime, str): Ending date of feeds to download (default today).
            Can be datetime or string (YYYYMMDD format with optional spacing).
        overwrite (bool): Flag for whether to overwrite any existing file (default False).
        use_requests (bool): Flag for whether to use requests or curl (default False == curl).
        overwrite_size_threshold (int): Existing files smaller than this will be re-downloaded.
        loop_sleep (int): Number of seconds to wait each loop after downloading file (default 1).

    Returns:
        tuple: output file path, return code
    """
    start_date = utilities.parse_date_input(start_date)
    end_date = utilities.parse_date_input(end_date or dt.datetime.today())
    loop_sleep = max(0.1, loop_sleep)

    _dls = []
    for i_date in utilities.iterate_dates(start_date, end_date, period="daily"):
        try:
            _dl = download_feed(
                i_date,
                overwrite=overwrite,
                use_requests=use_requests,
                overwrite_size_threshold=overwrite_size_threshold,
                sleep_after=loop_sleep,
            )
            _dls.append(_dl)
        except Exception as e:
            _logger.exception("Error downloading %d-%d-%d: %r", i_date.year, i_date.month, i_date.day, e)

    return _dls


def download_indexes_recursively(
    start_date,
    end_date=None,
    overwrite=False,
    use_requests=False,
    overwrite_size_threshold=8 * 1024,
    compressed=True,
    loop_sleep=1,
):
    """Download edgar quarterly compressed filing index files recursively from start to end.
    If `end_date` is `None`, default to today.
    Error downloads, or rate limited downloads are around 6kb,
    thus set the overwrite_size_threshold to larger than that
    to automatically re-download those error files.

    Args:
        start_date (datetime, str): Starting date of feeds to download. Can be datetime
            or string (YYYYMMDD format with optional spacing).
        end_date (None,datetime, str): Ending date of feeds to download. Default to today.
            Can be datetime or string (YYYYMMDD format with optional spacing).
        overwrite (bool): Flag for whether to overwrite any existing file. Default False).
        use_requests (bool): Flag for whether to use requests or curl. Default False == curl).
        overwrite_size_threshold (int): Existing files smaller than this will be re-downloaded. Default 8k.
        compressed (bool): Flag for whether compressed or uncompressed index files should be downloaded. Default True.
        loop_sleep (int): Number of seconds to wait each loop after downloading file (default 1).

    Returns:
        tuple: output file path, return code
    """
    start_date = utilities.parse_date_input(start_date)
    end_date = utilities.parse_date_input(end_date or dt.datetime.today())
    loop_sleep = max(0.1, loop_sleep)

    _dls = []
    for i_date in utilities.iterate_dates(start_date, end_date, period="quarterly"):
        edgar_url = get_index_url(i_date, compressed=compressed)
        index_cache = config.get_index_cache_path(i_date)

        try:
            _dl = download_from_edgar(
                edgar_url,
                index_cache,
                overwrite=overwrite,
                use_requests=use_requests,
                overwrite_size_threshold=overwrite_size_threshold,
                sleep_after=loop_sleep,
            )
            _dls.append(_dl)
        except Exception as e:
            _logger.exception("Error downloading %r: %r", i_date, e)

    return _dls



# -*- coding: utf-8 -*-
"""
General utilities for interacting with plaintext documents.

EDGAR document specification details: https://www.sec.gov/info/edgar/pdsdissemspec051310.pdf

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

import re
import logging
from collections import Counter

__logger = logging.getLogger(__name__)

_re_space = re.compile(r'\s')
_re_spaces = re.compile(r'\s+')
_re_spaces_2plus = re.compile(r'\s\s+')
_re_nl = re.compile(r'[\n\r]')
_re_nls = re.compile(r'[\n\r]+')
_re_number = re.compile(r'\b[\'"$-]*[\d.][\d.,]*\b')

def find_newlines(text, pos=0, endpos=-1):
    """
    Returns the list of the position of the _last_ set of newlines in a
    consecutive group. Includes the character at endposition.
    """
    if endpos <= pos:
        endpos = len(text)
    return list(x.end() for x in _re_nls.finditer(text, pos=pos, endpos=endpos))


def get_linestats(line, expected_line_length):
    r = {'line': line.rstrip()}
    r['line_strip'] = r['line'].lstrip()

    r['linelen'] = len(r['line'])
    r['textlen'] = len(r['line_strip'])
    r['space_left'] = r['linelen'] - r['textlen']
    r['space_right'] = expected_line_length - r['linelen']

    r['num_internal_spaces'] = len(_re_space.findall(r['line_strip']))
    r['num_internal_spacing'] = len(_re_spaces_2plus.findall(r['line_strip']))
    r['len_internal_spacing'] = sum(len(s) for s in
                                    _re_spaces_2plus.findall(r['line_strip']))
    r['num_tokens'] = len(r['line_strip'].split())
    r['num_numbers'] = sum(1 for _ in _re_number.finditer(r['line']))
    r['tok_sp_ratio'] = r['num_tokens'] / r['num_internal_spaces'] \
                        if r['num_internal_spaces'] else 0
    r['length_right_ratio'] = r['linelen'] / expected_line_length

    return r

def unwrap_plaintext(text, expected_line_length=None):
    splitlines = text.split('\n')

    if expected_line_length is None:
        aggline = Counter([len(l) for l in splitlines if l.strip()])

        if len(aggline) <= 1:
            return text

        mode_linelen = aggline.most_common(1)[0][0]
        max_linelen = max(aggline.keys())
        if mode_linelen / max_linelen < .95:
            __logger.debug("Poor line length format, mode: {} max: {} ({:.1f}%)"
                           .format(mode_linelen, max_linelen,
                                   100*mode_linelen/max_linelen))

        expected_line_length = mode_linelen

    joinlines = []
    stnext = get_linestats('', expected_line_length)

    for i,line in enumerate(splitlines):
        # get next line (which is stnext), and this loop's st is last's stnext
        stnext, st = get_linestats(line, expected_line_length), stnext

        newline = ((st['tok_sp_ratio'] < .9) |
                   # ^ Lots of space between tokens. Not normal text line.
                   (stnext['tok_sp_ratio'] < .75) |
                   # ^ Next line isn't normal line either
                   (st['length_right_ratio'] < .85))
                   # ^ Text doesn't go all the way to the end

        # If this line continues on next, remove left space from next line.
        if not newline:
            stnext['line'] = stnext['line_strip']

        # Remember, we're operating one line behind.
        joinlines.append(st['line'])
        joinlines.append('\n' if newline else ' ')
    # add last line.
    joinlines.append(stnext['line'])

    return ''.join(joinlines)

def get_para_bounds(text, position):
    r"""Gets location of nearest newlines surrounding `position` in `text`.

    :param text: Text to search for paragraphs in.
    :param position: Index of point around which newlines are searched.
    :return: Tuple containing (start, end) indexes of surrounding paragraph.
    :rtype: tuple

    >>> get_para_bounds('123\n567\n90', -1)
    (0, 3)
    >>> get_para_bounds('123\n567\n90', 5)
    (4, 7)
    >>> get_para_bounds('123\n567\n90', 100)
    (8, 10)
    >>> get_para_bounds('123\n567\n90', {})
    (0, 3)
    """
    start, end = None, None
    try:
        position = min(len(text), max(0, position))
    except TypeError:
        position = 0
    for i in range(2,10): # search 'backwards'
        for start in _re_nl.finditer(text, position-4**i, position):
            pass
        if start:
            break
    end = _re_nl.search(text, position)
    return start.end() if start else 0, end.start() if end else len(text)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*-

"""
Module for parsing HTML files.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

import re
import logging
from subprocess import Popen, PIPE

from . import plaintext
from ._html_encoding_lookup import html_ent_re_sub

__logger = logging.getLogger(__name__)

RE_HTML_TAGS = re.compile(r'</?(?:html|head|title|body|div|font|style|[apb]\b|tr|td|h\d)', re.I)

def convert_html_to_text(html_string, unwrap=True, document_width=150, force=False):
    """
    Get the plaintext version of an HTML string.
    If HTML, uses w3m linux program to parse into plain text.
    If `unwrap`, also unwraps paragraphs so each paragraph is on one line.
    TODO: come up with better unwrapping algorithm.

    :param string html_string: HTML or text document in a string
    :param bool unwrap: If True (default) call `plaintext.unwrap_plaintext` on text
    :param document_width: Expected width of lines in text (used for unwrapping, default=150).
    :param bool force: Skip checking whether text is HTML (# of valid HTML
        tags >= 3, default=False).

    :return: Plain text representation of file.
    :rtype: string
    """
    # If not an HTML file, just return the text.
    if (not force and
        (not html_string or len(RE_HTML_TAGS.findall(html_string, 0, 2000)) <= 3)):
        if unwrap:
            return plaintext.unwrap_plaintext(html_string, 80) # SGML is 80 chars wide
        return html_string

    text = html_ent_re_sub(html_string)

    p1 = Popen('w3m -T text/html -dump -cols {0} -no-graph'
                .format(document_width).split(),
                stdin=PIPE, stdout=PIPE)
    # Now send it the text on STDIN (like cat file.html | w3m)
    output = p1.communicate(input=text.encode())

    p1.stdout.close()  # Allow p1 to receive a SIGPIPE if p2 exits.
    try:
        p1.terminate()
    except ProcessLookupError:
        pass

    if output[-1]:
        __logger.warning(output[-1])

    if unwrap:
        return plaintext.unwrap_plaintext(output[0].decode(), document_width)

    return output[0].decode()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Use indices from EDGAR index files.

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

# Stdlib imports
import os
import re
import logging
# import datetime as dt

# 3rd party imports
import pandas as pd


# Module Imports
from pyedgar import config
from pyedgar.utilities.indices import IndexMaker as _IndexMaker


class EDGARIndex():
    """
    Class that opens EDGAR indices.
    Indices are always a pandas dataframe.
    """

    # Class variables
    simplify_col_names = None

    # Private class variables
    _logger = logging.getLogger(__name__)
    _simple_col_names = ('cik', 'name', 'form', 'filedate', 'accession')
    _raw_col_names = ('CIK', 'Company Name', 'Form Type', 'Date Filed', 'Accession')

    def __init__(self, simplify_col_names=True, force_download=False, use_tqdm=True):
        """
        Initialize the index making object.

        use_tqdm: flag for whether or not to wrap downloads in tqdm for progress monitoring
        """
        self.simplify_col_names = simplify_col_names
        self.IndexMaker = _IndexMaker(use_tqdm=use_tqdm)

        if force_download:
            self.IndexMaker.extract_indexes()


    @property
    def indices(self):
        """
        Search for all indices, return dict of all found.
        """
        return self.search_for_indices()

    def search_for_indices(self, fname_regex=None):
        """
        Search `config.INDEX_ROOT` for all index files, identified by
        `config.INDEX_EXTENSION`.

        Args:
            fname_regex (re): Regex to use to match against file names to find
                index files.

        Returns:
            dict: Dictionary of all found index files, as `{filename:fullpath,}`
        """
        idx_root = config.INDEX_ROOT
        indices = dict()

        if fname_regex is None:
            fname_regex = re.compile(r'\.{}$'.format(config.INDEX_EXTENSION), re.I)

        for _file in os.listdir(idx_root):
            matches = fname_regex.search(_file)
            self._logger.debug("Found %r which %r an index file.",
                              _file, 'matches' if matches else 'does not match')
            if matches:
                indices[_file] = os.path.join(idx_root, _file)

        return indices

    def get_index(self, index_name_or_path):
        # Assume they passed in the filing name, which is a key in self.indices
        try:
            return self.load_index(self.indices[index_name_or_path])
        except KeyError:
            pass

        # Next try to see if they omitted the extension
        try:
            try_lookup = '{}.{}'.format(index_name_or_path, config.INDEX_EXTENSION)
            return self.load_index(self.indices[try_lookup])
        except KeyError:
            pass

        # Last try to see if they omitted 'form' and the extension
        try:
            try_lookup = 'form_{}.{}'.format(index_name_or_path, config.INDEX_EXTENSION)
            return self.load_index(self.indices[try_lookup])
        except KeyError:
            pass

        # Then maybe they asked for a path
        return self.load_index(index_name_or_path)

    def load_index(self, index_path, override_sep=None):
        sep = config.INDEX_DELIMITER if override_sep is None else override_sep

        df = pd.read_csv(index_path, sep=sep)

        if self.simplify_col_names:
            df.rename(columns={k: v for k, v in
                               zip(self._raw_col_names, self._simple_col_names)},
                      index=str, inplace=True)
        if 'filedate' in df:
            df['filedate'] = pd.to_datetime(df['filedate'])

        return df

    def __getitem__(self, key):
        """
        Allow for dict-type lookup of indexes::

            idx = EDGARIndex()
            idx['10-K']
        """
        try:
            return self.get_index(key)
        except FileNotFoundError:
            raise KeyError("No index found at key {}. Indices found: {}"
                           .format(key, list(self.indices.keys())))



"""PyEDGAR is a python based package to interact with and download EDGAR filings.

Supports local caching of EDGAR database with the downloader module.
"""

import os
import re
from codecs import open

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the version from pyedgar/__init__.py
_version_re = re.compile(r'^__version__\s*=\s*[\'"](.*)[\'"]\s*$', re.M)

with open('pyedgar/__init__.py', 'r') as fh:
    version = _version_re.search(fh.read()).group(1).strip()

setup(
    name='pyedgar',
    version=version,
    description='Python interface to EDGAR filings.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords='SEC EDGAR filings',
    url='https://github.com/gaulinmp/pyedgar',
    author='Mac Gaulin',
    author_email='gaulinmp+pyedgar@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Topic :: Office/Business :: Financial :: Accounting',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=['pandas', 'requests'],
    extras_require={
        'dev': ['bs4', 'tqdm'],
        # 'test': ['coverage'],
    },
)



# -*- coding: utf-8 -*-
"""
pyEDGAR SEC data library.
=====================================

pyEDGAR is a general purpose library for all sorts of interactions with the SEC
data sources, primarily the EDGAR distribution system.

Files from the SEC reside at https://www.sec.gov/Archives/edgar/data/CIK/ACCESSION.txt

:copyright: © 2021 by Mac Gaulin
:license: MIT, see LICENSE for more details.
"""

__title__ = 'pyedgar'
__version__ = '0.1.7'
__version_info__ = tuple(int(i) for i in __version__.split("."))
__author__ = 'Mac Gaulin'
__license__ = 'MIT'
__copyright__ = 'Copyright 2021 Mac Gaulin'


# Include sub-modules
from pyedgar.filing import Filing
from pyedgar.index import EDGARIndex

# from pyedgar import utilities
# from pyedgar import exceptions
# from .exceptions import (InputTypeError, WrongFormType,
#                          NoFormTypeFound, NoCIKFound)

# __all__ = [edgarweb, forms, localstore, plaintext, #downloader,
#            InputTypeError, WrongFormType, NoFormTypeFound, NoCIKFound]
